"""Type aliases for type hints."""

from __future__ import annotations

from ._aliases import ArrayLike as ArrayLike
from ._aliases import BoundsTuple as BoundsTuple
from ._aliases import CellArrayLike as CellArrayLike
from ._aliases import CellsLike as CellsLike
from ._aliases import InteractionEventType as InteractionEventType
from ._aliases import MatrixLike as MatrixLike
from ._aliases import Number as Number
from ._aliases import RotationLike as RotationLike
from ._aliases import TransformLike as TransformLike
from ._aliases import VectorLike as VectorLike
from ._array_like import NumberType as NumberType
from ._array_like import NumpyArray as NumpyArray
from ._dataset_types import _DataObjectType as _DataObjectType
from ._dataset_types import _DataSetOrMultiBlockType as _DataSetOrMultiBlockType
from ._dataset_types import _DataSetType as _DataSetType
from ._dataset_types import _GridType as _GridType
from ._dataset_types import _PointGridType as _PointGridType
from ._dataset_types import _PointSetType as _PointSetType
